############################################################################
# David Konerding, LBNL
# See ../LBNLCopyright for copyright notice!
############################################################################


""" Template file for the service implementation.  pyGridWare
dispatches incoming messages for this service to the appropriate
methods """

import time, sys, os, tempfile

from twisted.internet import reactor
from distutils.sysconfig import get_python_lib
from PythonCLServiceTool.cat.underware.generated.resource.cat.PythonCLServiceTool_cat import ManagerHome

### Import the underware.
from PythonCLServiceTool.cat.underware.generated.services.cat.PythonCLServiceTool_cat import PythonCLServiceTool_catWSRF
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.lifetime.ResourceLifetime import \
    ImmediateResourceTermination, ScheduledResourceTermination
from pyGridWare.services.SubscriptionManager import \
    SubscriptionManager
from pyGridWare.utility.http import GetURLFromRequest

## Import required Twisted modules
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessDone, ProcessTerminated
from twisted.python.failure import Failure

from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import _get_type_definition as TED

## Parse the config file
from ConfigParser import ConfigParser
if os.getenv('PYCLST_BASE'):
    CONFIG_FILE=os.path.join(os.getenv('PYCLST_BASE'), "server-config.txt")
else:
    CONFIG_FILE=os.path.join(get_python_lib(), "PythonCLServiceTool", "cat", 'server-config.txt')
if not os.path.exists(CONFIG_FILE):
    raise RuntimeError, "Config file '%s' does not exist." % CONFIG_FILE
if not os.path.isfile(CONFIG_FILE):
    raise RuntimeError, "Config file '%s' is not a file." % CONFIG_FILE

CP = ConfigParser()
CP.read(CONFIG_FILE)
executable = CP.get('Application', 'executable')

## Convenience functions for getting classes for various schema types we use
_TARGETNAMESPACE = "urn:PythonCLServiceTool_cat_Schema"
optionArgumentType_Def = TED(_TARGETNAMESPACE, "optionArgumentType")
optionNoArgumentType_Def = TED(_TARGETNAMESPACE, "optionNoArgumentType")
positionalArgumentType_Def = TED(_TARGETNAMESPACE, "positionalArgumentType")


## Convert an executable and arguments (packed as XML message) into
## list of arguments suitable for passing to Twisted to start a new
## process
def gen_cl(executable, arguments):
    args = [executable]

    ## this happens if no arguments are passed (I orignally expected
    ## an empty list, but apparently this is how a sequence of Any
    ## behaves.  Should talk to Josh about this.)
    if arguments._arguments == None:
        return []
    
    for arg in arguments._arguments:
        if isinstance(arg.typecode, optionArgumentType_Def):
            if arg._isFile:
                t = tempfile.mktemp()
                open(t, "w").write(arg._value)
                args.extend([arg._option, t])
            else:
                args.extend([arg._option, arg._value])
        elif isinstance(arg.typecode, optionNoArgumentType_Def):
            args.append(arg._option)
        elif isinstance(arg.typecode, positionalArgumentType_Def):
            ## Convert client files to local files
            if arg._isFile:
                t = tempfile.mktemp()
                open(t, "w").write(arg._value)
                args.append(t)
            else:
                args.append(arg._value)
        else:
            ## TODO: validate better :)
            raise RuntimeError, "D'oh!"
    return args

## Implementation of Twisted's ProcessProtocol.  When the process
## state changes (or stdout/stderr arrives), apply the change to the
## appropriate resource property (this will trigger a notification).
## Also, stash all the stdout and stderr received in a list (to be
## agglomerated and sent at the end)
class PythonCLServiceTool_cat_ImplementationProcessProtocol(ProcessProtocol):
    def __init__(self, rp):
        self.rp = rp
        self.stdout = []
        self.stderr = []
        
    def connectionMade(self):
        self.rp.setstatus("connectionMade")
        self.transport.closeStdin()
        
    def inConnectionLost(self):
        self.rp.setstatus("inConnectionLost")

    def outConnectionLost(self):
        self.rp.setstatus("outConnectionLost")

    def errConnectionLost(self):
        self.rp.setstatus("errConnectionLost")

    def outReceived(self, data):
        self.stdout.append(data)
        self.rp.setstatus("outReceived, data: %s" % data)

    def errReceived(self, data):
        self.stderr.append(data)
        self.rp.setstatus("errReceived, data: %s" % data)

    def processEnded(self, status):
        if isinstance(status.value, ProcessDone):
            self.rp.setstatus("processEnded (terminated normally)")

        elif isinstance(status.value, ProcessTerminated):
            self.rp.setstatus("processEnded (terminated abnormally: status: %s)" % \
                               status.value.exitCode)
        
        self.rp.setreturnCode(status.value.exitCode)
        self.rp.setstdout("".join(self.stdout))
        self.rp.setstderr("".join(self.stderr))


## Implementation of the service.  Each of the messages for the service (as defined in the wsdl)
## is implemented with by a method of the same name.  
class PythonCLServiceTool_cat_Implementation(PythonCLServiceTool_catWSRF):

    def __init__(self, post='/services/PythonCLServiceTool_cat', **kw):
        PythonCLServiceTool_catWSRF.__init__(self, **kw)

    ## NOTE:Exceptions that occur in these functions are poorly
    ## propagated to the client, and can cause it to die an ugly,
    ## early death.  Must talk to Josh about this.


    ## Handle an incoming create request.  This creates a new resource
    ## context and an end point reference pointing to it.
    def wsa_create(self, ps, address, **kw):

        response = PythonCLServiceTool_catWSRF.wsa_create(self, ps, address, **kw)

        # Create a new Resource Context in which the job will be contained.
        ctx = self.NewResourceContextPythonCLServiceTool_catPort(GetURLFromRequest(kw['request']))

        rp = ctx.properties

        ## Set some initial RP values
        rp.setstatus('INITIAL')
        rp.setreturnCode(-1)
        ## next two lines to keep wsa_run error message from happening
        rp.setstdout('None')
        rp.setstderr('None')
        rp.setarguments(self.request._arguments)

        # Create EndPointReference and set in response
        epr = AddressingUtils.createEndpointReference(ctx)
        response._EndpointReference = epr
        return response

    ## Handle an incoming run request.  This causes an child process to be started by Twisted
    def wsa_run(self, ps, address, **kw):

        response = PythonCLServiceTool_catWSRF.wsa_run(self, ps, address)
        ctx = PythonCLServiceTool_catWSRF.GetResourceContext(ps, address)
        rp = ctx.properties

        ## Build the command line argument list
        args = gen_cl(executable, rp.getarguments())

        ## Start the child process.  It doesn't seem like we have a
        ## good way to tell "starting a process failed" from "the
        ## process failed when it was running" *-(
        reactor.spawnProcess(PythonCLServiceTool_cat_ImplementationProcessProtocol(rp), executable, args, env=os.environ)
        
        return response

