############################################################################
# David Konerding, LBNL
# See ../LBNLCopyright for copyright notice!
############################################################################

## Factored out command line parser code (used by client)

from optparse import OptionParser, OptionValueError
import sys, os

from PythonCLServiceTool.cat.underware.generated.types.cat import *

from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import _get_type_definition as TED

_TARGETNAMESPACE = "urn:PythonCLServiceTool_cat_Schema"

optionArgument_Dec = GED(_TARGETNAMESPACE, "optionArgument")
optionNoArgument_Dec = GED(_TARGETNAMESPACE, "optionNoArgument")
positionalArgument_Dec = GED(_TARGETNAMESPACE, "positionalArgument")

optionArgumentType_Def = TED(_TARGETNAMESPACE, "optionArgumentType")
optionNoArgumentType_Def = TED(_TARGETNAMESPACE, "optionNoArgumentType")
positionalArgumentType_Def = TED(_TARGETNAMESPACE, "positionalArgumentType")

## option parsing callback which finds positional args on the command
## line as well (optparse normally only supports positional args at
## the *end* of the command line.  But some apps and users intersperse
## positional and optional args)

## Note: there is a bit of unnecessary code duplication that should be
## factored out.

def testcallback(option, opt, value, parser):
    if value is not None:
        o = optionArgument_Dec()
        o.typecode = optionArgumentType_Def('arguments')
        o._option = opt
        ## Handle file transfer-- files with {} around their names are sent to the service.
        if value.startswith('{') and value.endswith('}'):
            fname = value[1:-1]
            if not os.path.exists(fname):
                raise OptionValueError, "Referenced filename %s does not exist in local filesystem." % fname
            o._value = open(fname).read()
            p._isFile = 1
        else:
            o._value = value
            o._isFile = 0
        parser.parsed_args.append(o)
    else:
        ## Collect data from an option argument (with no sub-argument)
        o = optionNoArgument_Dec()
        o.typecode = optionNoArgumentType_Def('arguments')
        o._option = opt
        parser.parsed_args.append(o)
    rargs = parser.rargs

    ## remove any trailing positional args
    while rargs:
        arg = rargs[0]
        if arg[0] == "-": break
        p = positionalArgument_Dec()
        p.typecode = positionalArgumentType_Def('arguments')
        if arg.startswith('{') and arg.endswith('}'):
            fname = arg[1:-1]
            if not os.path.exists(fname):
                raise OptionValueError, "Referenced filename %s does not exist in local filesystem." % fname
            p._value = open(fname).read()
            p._isFile = 1
        else:
            p._value = arg
            p._isFile = 0
        parser.parsed_args.append(p)
        del parser.rargs[0] 

def parse_args(args):
    usage='usage: %%prog [options] arg'
    parser = OptionParser()
    parser.parsed_args = [] 

    ## workaround for optparse: place any positional arguments at the start of the arg list into the 
    ## parsed_args list, and remove them from argument consideration.

    first = 0
    for arg in args:
        if arg[0] != "-":
            first += 1
            p = positionalArgument_Dec()
            p.typecode = positionalArgumentType_Def('arguments')
            if arg.startswith('{') and arg.endswith('}'):
                value = arg
                fname = value[1:-1]
                if not os.path.exists(fname):
                    raise OptionValueError, "Referenced filename %s does not exist in local filesystem." % fname
                p._value = open(value[1:-1]).read()
                p._isFile = 1
            else:
                p._value = arg
                p._isFile = 0
            parser.parsed_args.append(p)
        else: break

    parser.add_option('-n', dest='arg1', help='number', action='callback', callback=testcallback)
    parser.add_option('--version', dest='arg2', help='version', action='callback', callback=testcallback)
    (options, posargs) = parser.parse_args(args)

    return parser.parsed_args
