############################################################################
# David Konerding, LBNL
# See ../LBNLCopyright for copyright notice!
############################################################################


""" Template file for the client implementation """

import sys, time, urlparse, optparse, exceptions, socket, os
from ConfigParser import ConfigParser
from distutils.sysconfig import get_python_lib

from PythonCLServiceTool.NameTemplate.underware.generated.stubs.NameTemplate import SubscribeRequest, PythonCLServiceTool_NameTemplateLocator, createPythonCLServiceTool_NameTemplateRequest, RunRequestMessage, GetResourcePropertyRequest
from pyGridWare.notification.NotificationConsumer import NotificationConsumer
from pyGridWare.generated.services.basenotification.WS_BaseNotificationService import WS_BaseNotificationServiceWSRF as NCServiceInterface
import pyGridWare.patches

from ZSI.wstools.Namespaces import WSA200403

from twisted.python import log
from twisted.internet import reactor
from twisted.python import threadable
from twisted.internet.error import ConnectionRefusedError
threadable.init()

import cl_parser



DIALECT = "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"


## Parse config file settings.
if os.getenv('PYCLST_BASE'):
    CONFIG_FILE=os.path.join(os.getenv('PYCLST_BASE'), "client-config.txt")
else:
    CONFIG_FILE=os.path.join(get_python_lib(), "PythonCLServiceTool", "NameTemplate", 'client-config.txt')
if not os.path.exists(CONFIG_FILE):
    raise RuntimeError, "Config file '%s' does not exist." % CONFIG_FILE
if not os.path.isfile(CONFIG_FILE):
    raise RuntimeError, "Config file '%s' is not a file." % CONFIG_FILE
SEC_PATHS = 'Paths'
SEC_LOCATION = 'Location'
CP = ConfigParser()
CP.read(CONFIG_FILE)
HOST = CP.get(SEC_LOCATION, 'host')
PORT = CP.get(SEC_LOCATION, 'port')
DEBUG = int(CP.get('logging', 'debug'))

class Holder:
    pass


## Command line options (specific to the client, not the application).
## No longer used (because CL arguments are processed for the
## application) We'll have to figure out another way to do this,
## probably env vars.
## ###########################################################################
## op = optparse.OptionParser()
## op.add_option("-s", "--host",
##               action="store", dest="host", default=HOST,
##               help="location of service container")
## op.add_option("-p", "--port",
##               action="store", dest="port", default=PORT,
##               help="port of service container")
## (OPTIONS, ARGS) = op.parse_args()

OPTIONS=Holder()
OPTIONS.host = HOST
OPTIONS.port = PORT

## Holder for TopicExpression; used by notification system
class TopicExpression:
    '''
    '''
    def __init__(self):
        self._attrs = {}
        self._attrs['Dialect'] = DIALECT

    def setMixed(self, value):
        self._text = value

## Convenience function to retrieve a resource property
def GetResourceProperty(port, property):
    ## convert to unicode since the result we get back is...
    x = (u"urn:PythonCLServiceTool_NameTemplate_Schema", unicode(property))
    pyclass = GetResourcePropertyRequest
    request = pyclass(x)
    msg = port.GetResourceProperty(request)
    return msg._any[0]

## Class that handles incoming notifications (from the server).  When
## it receives the 'processEnded' status message, it collates all the
## stdout/stderr, gathers the return code, and stops the reactor.
class PythonCLServiceTool_NameTemplateNotificationConsumerService(NCServiceInterface):

    def __init__(self, post='/wsrf/services/NotificationConsumerService', **kw):
        NCServiceInterface.__init__(self)

    def wsa_Notify(self, ps, address, **kw):
        response = NCServiceInterface.wsa_Notify(self, ps, address)

        pyobj = self.request._NotificationMessage[0]
        pyobj = pyobj._Message._ResourcePropertyValueChangeNotification
        if pyobj._NewValue._any.startswith('processEnded'):

            returnCode = GetResourceProperty(self.port, 'returnCode')

            ## Currently, we only write stdout and stderr at the very end, by retrieving the value from the server.
            ## This has a number of performance (CPU, memory and network) issues associated with it.
            ## Probably we should pass notifications of stdout with a 'sequence number' (ala TCP) and let the
            ## client intelligently reassemble the messages when they arrive.  Every time something happens to stdout and stderr,
            ## the data gets sent as a notification anyway.
            stdout = GetResourceProperty(self.port, 'stdout')
            sys.stdout.write(stdout)
            sys.stdout.write("\n")
            stderr = GetResourceProperty(self.port, 'stderr')
            sys.stderr.write(stderr)
            sys.stderr.write("\n")
            self.master.returnCode = returnCode
            reactor.callWhenRunning(reactor.stop)
        return response
    
## Following two convenience methods are currently unused, but we
## should be calling 'Destroy' at the very least.
    
def Destroy(port):
    request = DestroyRequest()
    msg = port.Destroy(request)

def SetTerminationTime(port, interval):
    request = SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    msg = port.SetTerminationTime(request)



## "Master" class which invokes the remote service, sets up the
## notification receiver, and builds the argument XML data structure
class MasterClass:
    def __init__(self):
        ## Initialize a bunch of internal variables so we can check if
        ## they have been set later
        self.returnCode = None
        self.locator = None
        self.port = None
        self.epr = None
        self.service_port = None
        
        netloc = '%s:%s' % (OPTIONS.host, OPTIONS.port)
        path = CP.get(SEC_PATHS, 'PythonCLServiceTool_NameTemplate')
        urlTuple = ('http',netloc,path,'','','')
        self.url = urlparse.urlunparse(urlTuple)

    ## Send a create request message to the service.  Include the arguments
    def create(self):
        kw = {'url':self.url,
              'wsAddressURI':WSA200403.ADDRESS,
              }
        if DEBUG:
            kw['tracefile'] = sys.stdout

        try:
            args = cl_parser.parse_args(sys.argv[1:])

        ## this except handles invalid command lines (unexpected options) (raised by OptionParser)
        except exceptions.SystemExit, exitcode:
            ## negate the error code to reduce the likelihood of
            ## stomping a real application error code
            self.returnCode = -exitcode.code
            reactor.callWhenRunning(reactor.stop)
            return

        ## this exception handles invalid command lines (referenced
        ## local file to be shipped to the server doesn't exist
        ## locally)
        except optparse.OptionValueError, what:
            log.err(what)
            reactor.callWhenRunning(reactor.stop)
            return

        ## Build the create request
        self.locator = PythonCLServiceTool_NameTemplateLocator()
        self.port = self.locator.getPythonCLServiceTool_NameTemplatePortType(**kw)

        request = createPythonCLServiceTool_NameTemplateRequest()
        request._arguments = Holder()
        request._arguments._arguments = []

        for arg in args:
            request._arguments._arguments.append(arg)

        try:
            response = self.port.create(request)
        ## TODO: merge all three exceptions using exception tuple format
        except socket.gaierror, what:
            log.err("Failed to connect to remote server, url=%s, message=%s" % (self.url, what))
            self.returnCode = -1
            reactor.callWhenRunning(reactor.stop)
            return
        except socket.error, what:
            log.err("Failed to connect to remote server, url=%s, message=%s" % (self.url, what))
            self.returnCode = -1
            reactor.callWhenRunning(reactor.stop)
            return
        except ConnectionRefusedError, what:
            log.err("Failed to connect to remote server, url=%s, message=%s" % (self.url, what))
            self.returnCode = -1
            reactor.callWhenRunning(reactor.stop)
            return
            
        ## attempt to catch internal error:

        ## File
        ## "/home/portnoy/u5/dek/sw/rh9/Python-2.4/lib/python2.4/site-packages/pyGridWare/generated/stubs/PythonCLServiceTool_ls.py",
        ## line 42, in create raise TypeError, "%s incorrect response
        ## type" % (response.__class__) exceptions.TypeError:
        ## twisted.python.failure.Failure incorrect response type.
        ## (Must talk to Josh.)
        except exceptions.TypeError:
            log.err("Exception -- unsupported response from server.  Check server logs")
            self.returnCode = -1
            reactor.callWhenRunning(reactor.stop)
            return

        ## Stash away the EPR provided by the service (of the specific
        ## service instance we will communicate with)
        self.epr = response._EndpointReference

        # Stash away a port associated with the EPR that we will make invocations upon
        self.service_port = self.locator.getPythonCLServiceTool_NameTemplatePortType(portAddress=self.url, \
                                                                               endPointReference=self.epr, \
                                                                               **kw)
    ## Start the notification consumer.
    def consumer(self):
        self.consumer = NotificationConsumer(notificationClass=PythonCLServiceTool_NameTemplateNotificationConsumerService)
        if not self.service_port:
            ## Hmm; this shouldn't really happen.  Maybe use an assert here instead.
            raise RuntimeError, "Consumer: not ready because service port is not available!"
        self.consumer.consumer.port = self.service_port
        self.consumer.consumer.master = self
        self.consumer.start()

    ## Queue up events-- first, send a subscription (should only be
    ## done *after* the consumer has started), second, send the Run request message to the service
    def enqueue(self):
        if not self.service_port:
            ## Hmm; this shouldn't really happen.  Maybe use an assert here instead.
            raise RuntimeError, "Enqueue: not ready because service port is not available!"

        reactor.callWhenRunning(self.Subscribe)
        reactor.callWhenRunning(self.Run)

    ## Subscribe to notifications related to our job's progress.
    def Subscribe(self):
        """
        port -- Subscription Client
        consumer -- Notification Consumer
        Need to specify the prefix-namespace mapping in the
        XML instance to deal with QNames passed via the 
        TopicExpression mixed text content.
        """

        request = SubscribeRequest()
        self.consumer_epr = self.consumer.getEndPointReference()
        request._ConsumerReference = self.consumer_epr
        request._TopicExpression = TopicExpression()
        request._TopicExpression.setMixed("gsr:status")
        request._UseNotify = True
        #request._Precondition = None
        #request._Selector = None
        #request._SubscriptionPolicy = None
        #request._InitialTerminationTime = None
        self.service_port.binding.nsdict['gsr'] = "urn:PythonCLServiceTool_NameTemplate"
        msg = self.service_port.Subscribe(request)

    ## Send a run request to the service endpoint, causing the service
    ## to start running the job.
    def Run(self):
        request = RunRequestMessage()
        msg = self.service_port.run(request)

def main(c):
    c.create()

    if not c.returnCode:
        c.consumer()
        c.enqueue()

if __name__ == '__main__':
    c = MasterClass()

    ## Delay invocation until the Twisted reactor is fully running
    ## (otherwise there is a possibility of losing some early
    ## notification messages, because the EPR won't be ready to
    ## receive yet)
    reactor.callWhenRunning(main, c)
    if DEBUG:
        log.startLogging(sys.stdout, 0)
    reactor.run()

    # bugs in the server can cause the process to not complete (aborting received a response to a SOAP request);
    # we catch that here
    if c.returnCode is None:
        raise RuntimeError, "Unexpected error: no return code received.  Up the log level to DEBUG and try again."
    
    sys.exit(c.returnCode)
